'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.InternalNode = exports.Leaf = exports.Node = exports.DELETION_STRATEGIES = undefined;

var _binarysearch = require('./binarysearch');

var _binarysearch2 = _interopRequireDefault(_binarysearch);

var _constants = require('./constants');

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var binSearch = _binarysearch2.default.eq;

var internalInsertKeyAt = function internalInsertKeyAt(cmp, key, keys) {
    return _binarysearch2.default.gte(keys, key, cmp);
};

function Node(_opts) {
    var opts = _opts || {};
    this.keys = opts.keys || [];
    this.children = opts.children || [];
    this.order = _constants.ORDER;
    this.ownerID = opts.ownerID;
}

Object.defineProperty(Node.prototype, 'size', {
    enumerable: true,
    get: function get() {
        return this.children.length;
    }
});

(0, _utils.extend)(Node.prototype, {
    satisfiesMinChildren: function satisfiesMinChildren() {
        return this.children.length >= this.minChildren;
    },
    satisfiesMaxChildren: function satisfiesMaxChildren() {
        return this.children.length <= this.maxChildren;
    },
    tail: function tail(ownerID) {
        return new this.constructor({
            order: this.order,
            keys: (0, _utils.tail)(ownerID, this.keys),
            children: (0, _utils.tail)(ownerID, this.children),
            ownerID: ownerID
        });
    },
    init: function init(ownerID) {
        return new this.constructor({
            order: this.order,
            keys: (0, _utils.init)(ownerID, this.keys),
            children: (0, _utils.init)(ownerID, this.children),
            ownerID: ownerID
        });
    },
    shouldSplit: function shouldSplit() {
        return !this.satisfiesMaxChildren();
    }
});

function Leaf(opts) {
    Node.call(this, opts);
}

Leaf.prototype = Object.create(Node.prototype);
Leaf.prototype.constructor = Leaf;

(0, _utils.extend)(Leaf.prototype, {
    maxChildren: _constants.LEAF_MAX_CHILDREN,

    minChildren: _constants.LEAF_MIN_CHILDREN,

    delete: function _delete(cmp, ownerID, didChange, key) {
        var idx = binSearch(this.keys, key, cmp);
        // Key was not found. No modifications needed.
        if (idx === -1) {
            return this;
        }

        (0, _utils.setRef)(didChange);

        var newLeaf = void 0;
        var newKeys = (0, _utils.withoutIdx)(ownerID, idx, this.keys);
        var newChildren = (0, _utils.withoutIdx)(ownerID, idx, this.children);

        if ((0, _utils.canMutate)(this, ownerID)) {
            newLeaf = this;
            newLeaf.keys = newKeys;
            newLeaf.children = newChildren;
        } else {
            newLeaf = new Leaf({
                order: this.order,
                keys: (0, _utils.withoutIdx)(ownerID, idx, this.keys),
                children: (0, _utils.withoutIdx)(ownerID, idx, this.children),
                ownerID: ownerID
            });
        }

        return newLeaf;
    },


    /**
     * Returns a new leaf with entries merged
     * from `otherLeaf`. The keys in `otherLeaf`
     * must be higher than in the current leaf,
     * i.e. `otherLeaf` must reside to the right of
     * this leaf in the tree.
     * @param  {Leaf} otherLeaf - other leaf to merge
     * @return {Leaf} a new Leaf with entries merged.
     */
    merge: function merge(otherLeaf) {
        return new Leaf({
            order: this.order,
            keys: this.keys.concat(otherLeaf.keys),
            children: this.children.concat(otherLeaf.children)
        });
    },
    idxForKey: function idxForKey(cmp, key) {
        return _binarysearch2.default.gte(this.keys, key, cmp);
    },
    insert: function insert(cmp, ownerID, didChange, key, value) {
        var idx = this.idxForKey(cmp, key);
        var alreadyHasKey = this.keys[idx] === key;

        var newKeys = void 0;
        var newChildren = void 0;

        if (!alreadyHasKey) {
            newKeys = (0, _utils.insert)(ownerID, idx, key, this.keys);
            newChildren = (0, _utils.insert)(ownerID, idx, value, this.children);
        } else {
            var existingValue = this.children[idx];
            if (existingValue === value) {
                return this;
            }

            newKeys = (0, _utils.set)(ownerID, idx, key, this.keys);
            newChildren = (0, _utils.set)(ownerID, idx, value, this.children);
        }

        (0, _utils.setRef)(didChange);

        var newLeaf = void 0;
        if ((0, _utils.canMutate)(this, ownerID)) {
            newLeaf = this;
            this.keys = newKeys;
            this.children = newChildren;
        } else {
            newLeaf = new Leaf({
                order: this.order,
                keys: newKeys,
                children: newChildren,
                ownerID: ownerID
            });
        }

        return newLeaf.shouldSplit() ? newLeaf.split(ownerID) : newLeaf;
    },
    split: function split(ownerID) {
        var cutoff = (0, _utils.median)(this.keys.length);
        var smallestKeyInOther = this.keys[cutoff];

        var keypair = (0, _utils.splitAt)(ownerID, cutoff, this.keys);
        var thisKeys = keypair[0];
        var otherKeys = keypair[1];

        var childrenPair = (0, _utils.splitAt)(ownerID, cutoff, this.children);
        var thisChildren = childrenPair[0];
        var otherChildren = childrenPair[1];

        var other = new Leaf({
            order: this.order,
            keys: otherKeys,
            children: otherChildren,
            ownerID: ownerID
        });

        var thisSplit = void 0;
        if ((0, _utils.canMutate)(this, ownerID)) {
            thisSplit = this;
            thisSplit.keys = thisKeys;
            thisSplit.children = thisChildren;
        } else {
            thisSplit = new Leaf({
                order: this.order,
                keys: thisKeys,
                children: thisChildren,
                ownerID: ownerID
            });
        }

        return [smallestKeyInOther, thisSplit, other];
    },
    smallestKey: function smallestKey() {
        return this.keys[0];
    },
    stealFirstKeyFrom: function stealFirstKeyFrom(ownerID, rightSibling) {
        var stolenKey = rightSibling.keys[0];
        var stolenValue = rightSibling.children[0];

        // Note: mutative. If we're stealing a key,
        // it means this node was created during
        // the edit and can be mutated.
        this.keys = this.keys.concat(stolenKey);
        this.children = this.children.concat(stolenValue);

        var siblingWithoutFirstKey = rightSibling.tail(ownerID);
        return [this, siblingWithoutFirstKey];
    },
    giveLastKeyTo: function giveLastKeyTo(ownerID, rightSibling) {
        var keyToGive = this.keys[this.keys.length - 1];
        var valueToGive = this.children[this.children.length - 1];

        // Note: mutative. If we're giving a key,
        // it means the sibling node was created during
        // the edit and can be mutated.
        rightSibling.keys = (0, _utils.unshift)(ownerID, keyToGive, rightSibling.keys);
        rightSibling.children = (0, _utils.unshift)(ownerID, valueToGive, rightSibling.children);

        var thisWithoutLastKey = this.init(ownerID);
        return [thisWithoutLastKey, rightSibling];
    }
});

function InternalNode(opts) {
    Node.call(this, opts);
}

var REPLACE = 'REPLACE';
var STEAL_KEY_FROM_LEFT = 'STEAL_KEY_FROM_LEFT';
var STEAL_KEY_FROM_RIGHT = 'STEAL_KEY_FROM_RIGHT';
var MERGE = 'MERGE';

var DELETION_STRATEGIES = exports.DELETION_STRATEGIES = {
    STEAL_KEY_FROM_LEFT: STEAL_KEY_FROM_LEFT,
    STEAL_KEY_FROM_RIGHT: STEAL_KEY_FROM_RIGHT,
    MERGE: MERGE
};

InternalNode.prototype = Object.create(Node.prototype);
InternalNode.prototype.constructor = InternalNode;

(0, _utils.extend)(InternalNode.prototype, {
    maxChildren: _constants.INTERNAL_MAX_CHILDREN,

    minChildren: _constants.INTERNAL_MIN_CHILDREN,

    merge: function merge(otherNode) {
        // We have ak + bk keys, and ac + bc children.
        // In a node, ac - ak === 1 (one less key than children).
        // Because (ac + bc) - (ak + bk) === 2, we need to add a key.
        // The key is the smallest key on the right node subtree.
        var toConcat = (0, _utils.unshift)(null, otherNode.smallestKey(), otherNode.keys);
        var newNode = new InternalNode({
            order: this.order,
            keys: this.keys.concat(toConcat),
            children: this.children.concat(otherNode.children)
        });
        return newNode;
    },
    chooseComplexDeletionStrategy: function chooseComplexDeletionStrategy(childIdx, newChild) {
        // When we need to merge, steal or give keys to
        // siblings -- this method returns the operation parameters,
        // which are: the left node, the right node, the index of
        // the left node in this nodes `children` array, and the
        // strategy to use (merge, steal from right, steal from left).

        if (newChild.satisfiesMinChildren()) {
            return {
                strategy: REPLACE
            };
        }

        var hasRightSibling = childIdx + 1 < this.children.length;
        var hasLeftSibling = childIdx - 1 >= 0;

        var isLeaf = newChild.constructor === Leaf;

        var nullSibling = { size: 0 };

        var rightSibling = hasRightSibling ? this.children[childIdx + 1] : nullSibling;

        var leftSibling = hasLeftSibling ? this.children[childIdx - 1] : nullSibling;

        var minChildren = isLeaf ? _constants.LEAF_MIN_CHILDREN : _constants.INTERNAL_MIN_CHILDREN;

        var strategy = void 0;
        if (rightSibling.size >= leftSibling.size) {
            if (rightSibling.size <= minChildren) {
                strategy = MERGE;
            } else {
                strategy = STEAL_KEY_FROM_RIGHT;
            }

            return {
                leftNode: newChild,
                rightNode: rightSibling,
                leftNodeIdx: childIdx,
                strategy: strategy
            };
        } else {
            if (leftSibling.size <= minChildren) {
                strategy = MERGE;
            } else {
                strategy = STEAL_KEY_FROM_LEFT;
            }
            return {
                leftNode: leftSibling,
                rightNode: newChild,
                leftNodeIdx: childIdx - 1,
                strategy: strategy
            };
        }
    },
    delete: function _delete(cmp, ownerID, didChange, key) {
        var childIdx = this.childIdxForKey(cmp, key);
        var origChild = this.children[childIdx];
        var child = origChild.delete(cmp, ownerID, didChange, key);

        if (!(0, _utils.isSet)(didChange)) return this;

        var strategyInfo = this.chooseComplexDeletionStrategy(childIdx, child);
        var strategy = strategyInfo.strategy;

        if (strategy === REPLACE) {
            return this.withReplacedChildren(ownerID, childIdx, [child]);
        }

        var leftNode = strategyInfo.leftNode;
        var rightNode = strategyInfo.rightNode;
        var leftNodeIdx = strategyInfo.leftNodeIdx;


        if (strategy === MERGE) {
            return this.withMergedChildren(ownerID, leftNodeIdx, leftNode, rightNode);
        }

        var newLeftNode = void 0;
        var newRightNode = void 0;
        if (strategy === STEAL_KEY_FROM_RIGHT) {
            var newNodes = leftNode.stealFirstKeyFrom(ownerID, rightNode);
            newLeftNode = newNodes[0];
            newRightNode = newNodes[1];
        } else if (strategy === STEAL_KEY_FROM_LEFT) {
            var _newNodes = leftNode.giveLastKeyTo(ownerID, rightNode);
            newLeftNode = _newNodes[0];
            newRightNode = _newNodes[1];
        }

        var withReplacedChildren = this.withReplacedChildren(ownerID, leftNodeIdx, [newLeftNode, newRightNode]);

        // Key must be updated because of the child key theft.
        var keyIdxToReplace = leftNodeIdx;
        var newKey = newRightNode.smallestKey();

        withReplacedChildren.keys = (0, _utils.set)(ownerID, keyIdxToReplace, newKey, withReplacedChildren.keys);
        return withReplacedChildren;
    },
    withMergedChildren: function withMergedChildren(ownerID, leftChildIdx, leftNode, rightNode) {
        var mergedChild = leftNode.merge(rightNode);

        var keyIdxToPop = leftChildIdx;

        var newKeys = (0, _utils.withoutIdx)(ownerID, keyIdxToPop, this.keys);

        var areLeftmostNodes = leftChildIdx === 0;

        if (!areLeftmostNodes) {
            newKeys[leftChildIdx - 1] = mergedChild.smallestKey();
        }

        var newChildren = (0, _utils.arrayClone)(this.children);
        newChildren.splice(leftChildIdx, 1);
        newChildren[leftChildIdx] = mergedChild;

        var withReplacedChild = new InternalNode({
            order: this.order,
            keys: newKeys,
            children: newChildren,
            ownerID: ownerID
        });
        return withReplacedChild;
    },
    childIdxForKey: function childIdxForKey(cmp, key) {
        return _binarysearch2.default.lte(this.keys, key, cmp) + 1;
    },
    stealFirstKeyFrom: function stealFirstKeyFrom(ownerID, rightSibling) {
        // Note that we mutate `this`.
        // Whenever we're stealing a first key,
        // it means this node was created during
        // the current update and can be mutated.
        var stolenValue = rightSibling.children[0];

        this.keys = this.keys.concat(rightSibling.smallestKey());
        this.children = this.children.concat(stolenValue);

        return [this, rightSibling.tail(ownerID)];
    },
    giveLastKeyTo: function giveLastKeyTo(ownerID, rightSibling) {
        // Steal last key-value pair from left node
        var stolenValue = (0, _utils.last)(this.children);
        rightSibling.keys = (0, _utils.unshift)(ownerID, rightSibling.smallestKey(), rightSibling.keys);
        rightSibling.children = (0, _utils.unshift)(ownerID, stolenValue, rightSibling.children);

        return [this.init(ownerID), rightSibling];
    },
    withReplacedChildren: function withReplacedChildren(ownerID, idx, newChildren) {
        var replaced = (0, _utils.canMutate)(this.children, ownerID) ? this.children : (0, _utils.arrayClone)(this.children);

        for (var i = 0; i < newChildren.length; i++) {
            replaced[idx + i] = newChildren[i];
        }

        if ((0, _utils.canMutate)(this, ownerID)) {
            this.children = replaced;
            return this;
        }

        return new InternalNode({
            order: this.order,
            keys: this.keys,
            children: replaced,
            ownerID: ownerID
        });
    },
    smallestKey: function smallestKey() {
        var curr = this;
        while (curr.constructor !== Leaf) {
            curr = curr.children[0];
        }
        return curr.keys[0];
    },
    split: function split(ownerID) {
        var medianIdx = (0, _utils.median)(this.keys.length) - 1;

        var splitArr = (0, _utils.takeIdxAndSplit)(ownerID, medianIdx, this.keys);
        var thisKeys = splitArr[0];
        var medianKey = splitArr[1];
        var otherKeys = splitArr[2];

        var childrenPair = (0, _utils.splitAt)(ownerID, medianIdx + 1, this.children);
        var thisChildren = childrenPair[0];
        var otherChildren = childrenPair[1];

        var thisNode = void 0;
        if ((0, _utils.canMutate)(this, ownerID)) {
            thisNode = this;
            thisNode.keys = thisKeys;
            thisNode.children = thisChildren;
        } else {
            thisNode = new InternalNode({
                order: this.order,
                keys: thisKeys,
                children: thisChildren,
                ownerID: ownerID
            });
        }

        var otherNode = new InternalNode({
            order: this.order,
            keys: otherKeys,
            children: otherChildren,
            ownerID: ownerID
        });

        return [medianKey, thisNode, otherNode];
    },
    withSplitChild: function withSplitChild(cmp, ownerID, newKey, splitChild, newChild) {
        var insertNewKeyAt = internalInsertKeyAt(cmp, newKey, this.keys);
        var newKeys = (0, _utils.insert)(ownerID, insertNewKeyAt, newKey, this.keys);

        var newChildren = (0, _utils.insert)(ownerID, insertNewKeyAt + 1, newChild, this.children);
        // Replace the original child with the split one.
        newChildren[insertNewKeyAt] = splitChild;

        if ((0, _utils.canMutate)(this, ownerID)) {
            this.keys = newKeys;
            this.children = newChildren;
            return this;
        }

        return new InternalNode({
            order: this.order,
            keys: newKeys,
            children: newChildren,
            ownerID: ownerID
        });
    },
    insert: function insert(cmp, ownerID, didChange, key, value) {
        var childIdx = this.childIdxForKey(cmp, key);
        var child = this.children[childIdx];

        var newChild = child.insert(cmp, ownerID, didChange, key, value);

        if (!(0, _utils.isSet)(didChange)) return this;

        // Got new child.

        // Child was split on insertion.
        if (newChild.length === 3) {
            var splitArr = newChild;
            var medianKey = splitArr[0];
            var splitChild = splitArr[1];
            var _newChild = splitArr[2];

            var withSplitChild = this.withSplitChild(cmp, ownerID, medianKey, splitChild, _newChild);
            return withSplitChild.shouldSplit() ? withSplitChild.split(ownerID) : withSplitChild;
        }

        return this.withReplacedChildren(ownerID, childIdx, [newChild]);
    }
});

exports.Node = Node;
exports.Leaf = Leaf;
exports.InternalNode = InternalNode;