'use strict';

var _chai = require('chai');

var _chai2 = _interopRequireDefault(_chai);

var _sinonChai = require('sinon-chai');

var _sinonChai2 = _interopRequireDefault(_sinonChai);

var _flatten = require('ramda/src/flatten');

var _flatten2 = _interopRequireDefault(_flatten);

var _utils = require('../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

_chai2.default.use(_sinonChai2.default);
var expect = _chai2.default.expect;


function assertArrayShallowEquals(a, b) {
    expect(a).to.have.lengthOf(b.length);
    for (var i = 0; i < a.length; i++) {
        expect(a[i]).to.equal(b[i]);
    }
}

describe('Utils', function () {
    it('bounded chunk', function () {
        var size = 102;
        var data = new Array(size);
        var min = 5;
        var max = 10;
        var chunks = (0, _utils.boundedChunk)(min, max, data);
        chunks.forEach(function (chunk) {
            expect(chunk.length).to.be.at.least(min);
            expect(chunk.length).to.be.at.most(max);
        });
        expect((0, _flatten2.default)(chunks).length).to.equal(size);
    });

    it('slice', function () {
        var arr = [0, 1, 2, 3, 4];
        var result = (0, _utils.slice)(null, 1, 3, arr);
        assertArrayShallowEquals(result, [1, 2]);
    });

    it('slice with mutations', function () {
        var ownerID = (0, _utils.makeOwnerID)();
        var arr = (0, _utils.tagOwnerID)([0, 1, 2, 3, 4], ownerID);
        var result = (0, _utils.slice)(ownerID, 1, 3, arr);
        expect(result).to.equal(arr);

        assertArrayShallowEquals(result, [1, 2]);
    });

    it('fastMap', function () {
        var arr = [0, 1, 2, 3];
        var result = (0, _utils.fastMap)(function (num) {
            return num.toString();
        }, arr);
        expect(result).to.deep.equal(['0', '1', '2', '3']);
    });

    it('insert', function () {
        var arr = [0, 1, 3, 4];
        var result = (0, _utils.insert)(null, 2, 2, arr);
        expect(result).to.deep.equal([0, 1, 2, 3, 4]);
    });

    it('insert with mutations', function () {
        var id = {};
        var arr = [0, 1, 3, 4];
        (0, _utils.tagOwnerID)(arr, id);
        var result = (0, _utils.insert)(id, 2, 2, arr);
        expect(result).to.equal(arr);
        assertArrayShallowEquals(result, [0, 1, 2, 3, 4]);
    });

    it('withoutIdx', function () {
        var arr = [0, 1, 1, 2, 3];
        var result = (0, _utils.withoutIdx)(null, 2, arr);
        assertArrayShallowEquals(result, [0, 1, 2, 3]);
    });

    it('withoutIdx with mutations', function () {
        var ownerID = (0, _utils.makeOwnerID)();
        var arr = (0, _utils.tagOwnerID)([0, 1, 1, 2, 3], ownerID);
        var result = (0, _utils.withoutIdx)(ownerID, 2, arr);
        expect(result).to.equal(arr);
        assertArrayShallowEquals(result, [0, 1, 2, 3]);
    });

    it('splitAt', function () {
        var arr = [0, 1, 2, 3];
        var result = (0, _utils.splitAt)(null, 2, arr);

        var left = result[0];
        var right = result[1];

        assertArrayShallowEquals(left, [0, 1]);
        assertArrayShallowEquals(right, [2, 3]);
    });

    it('unshift', function () {
        var arr = [1, 2, 3];
        var result = (0, _utils.unshift)(null, 0, arr);
        expect(result).to.deep.equal([0, 1, 2, 3]);
    });

    it('unshift with mutations', function () {
        var ownerID = {};
        var arr = (0, _utils.tagOwnerID)([1, 2, 3], ownerID);
        var result = (0, _utils.unshift)(ownerID, 0, arr);
        expect(result).to.equal(arr);
        assertArrayShallowEquals(result, [0, 1, 2, 3]);
    });

    it('takeIdxAndSplit', function () {
        var arr = [1, 2, 3, 4, 5];
        var result = (0, _utils.takeIdxAndSplit)(null, 2, arr);

        var leftResult = result[0];
        var takenIdx = result[1];
        var rightResult = result[2];

        assertArrayShallowEquals(leftResult, [1, 2]);
        expect(takenIdx).to.equal(3);
        assertArrayShallowEquals(rightResult, [4, 5]);
    });

    it('set', function () {
        var arr = [1, 999, 3];
        var result = (0, _utils.set)(null, 1, 2, arr);
        expect(result).not.to.equal(arr);
        assertArrayShallowEquals(result, [1, 2, 3]);
    });

    it('set with mutations', function () {
        var ownerID = (0, _utils.makeOwnerID)();
        var arr = (0, _utils.tagOwnerID)([1, 999, 3], ownerID);
        var result = (0, _utils.set)(ownerID, 1, 2, arr);
        expect(result).to.equal(arr);
        assertArrayShallowEquals(result, [1, 2, 3]);
    });
});